AUXVAR = 'auxvar_tensor'
ITERVAR = 'iteration_variable'
THRESHOLD_TENSOR = "ptq_threshold_tensor"
GPTQ_ITER = "_gptq_iter"
AUXSHIFT = '_shift'
TEMP = '_temp'
WEIGHTS_QUANTIZATION_PARAMS = 'weights_quantization_params'
PTQ_MIN_RANGE = "_min_range"
PTQ_MAX_RANGE = "_max_range"
